
; LZdpack - Lempel-Ziv pack routine
; LET bc=i=(l=(o+USR LZdpack))
; i=input address, l=input length, o=output address
; LZdpack=org+36

        org  64500

MAXSIZE:equ  63
WWIDTH: equ  255             ;window width
STKEND: equ  23653
FPTOBC: equ  $2da2           ;from calculator stack: BC

IBEG:   defw 00000           ;*entered: input start
MAXLEN: defb MAXSIZE         ;maximum match length
MCWORK: defs 33              ;start of working storage

PTMP:   equ  MCWORK          ;00 window match start address (temporary)
PBEG:   equ  MCWORK+2        ;02 window match start address (final)
PLEN:   equ  PBEG+2          ;04 window match length
OBYTES: equ  PLEN+2          ;06 encoded bytes output
CPOS:   equ  OBYTES+2        ;08 coding position
CNXT:   equ  CPOS+2          ;10 lookahead pointer
LEND:   equ  CNXT+2          ;12 encoding end
LPOS:   equ  LEND+2          ;14 *entered: output start
WINDIFF:equ  LPOS+2          ;16 (WINZ-PINA+1)-(WWIDTH) flag: 0=equal, else=not equal
PWIDTHF:equ  LPOS+3          ;17 window width (adjusted) flag: 0=zero, else=not zero
WINA:   equ  LPOS+4          ;18 LH window boundary
WINZ:   equ  WINA+2          ;20 RH window boundary
PINA:   equ  WINZ+2          ;22 LH window boundary (adjusted)

ILEN:   equ  MCWORK+24 ;27       ;*entered: input length (256-10000)
IEND:   equ  MCWORK+26 ;29       ;=calculated: input end (IBEG+ILEN)
ILAST:  equ  MCWORK+28 ;31       ;=calculated: input last byte address (IEND-1)
CODSEQ: equ  MCWORK+30 ;34       ;coding flag byte sequence: 8->1
CODADD: equ  MCWORK+31 ;35       ;coding flag byte address

;ENCODING
; Coding flag byte specifies length of next 8 entries.
; Coding entries may be:
; 0 matches: 1 byte  : cccccccc           : cchar
; 1 match:   1 byte  : mmmmmmmm           : mchar
;2+ matches: 2 bytes : llllllll oooooooo  : length+offset

        ;ENCODING SETUP
SETUPE: call FPTOBC
        inc  bc              ;leave 2 bytes for (OBYTES)
        inc  bc
        ld   (CODADD),bc     ;=coding flag byte address
        xor  a
        ld   (bc),a
        inc  bc              ; and 1 byte for 1st coding flag
        ld   (LPOS),bc       ;=start of output
        call FPTOBC
        ld   (ILEN),bc       ;=input length
        push bc           ;1>
        call FPTOBC
        ld   (IBEG),bc       ;=start of input
        ld   hl,(STKEND)
        ld   de,15
        add  hl,de
        ld   (STKEND),hl     ;restore (STKEND)

        ld   h,b
        ld   l,c             ;(IBEG)=start of input
        ld   (CPOS),hl       ;(CPOS)=coding position
        pop  bc           ;1<;(ILEN)
        push hl           ;1>
        add  hl,bc
        ld   (IEND),hl       ;=end of input (1st byte after)
        dec  hl
        ld   (ILAST),hl      ;input last byte address (IEND-1)
        inc  hl
        ld   de,1
        ld   (OBYTES),de     ;(OBYTES) =encoded bytes output
        ld   ix,MCWORK       ;
        ld   (ix+30),8       ;(CODSEQ)=coding flag byte sequence
        pop  hl           ;1<

        ld   bc,WWIDTH       ;WWIDTH=window width (one byte)
        sbc  hl,bc           ;HL=IBEG-WWIDTH
        ld   (WINA),hl       ;WINA,WINZ=left,right window boundaries
        ld   hl,(IBEG)
        dec  hl
        ld   (WINZ),hl

        exx
        push hl              ;save H'L' (holds calculator RET address)

        xor  a
        ld   (PLEN),a
        ld   (PWIDTHF),a

        ;START OF THE ENCODING LOOP
ENLOOP: ld   hl,(WINA)       ;#3
        ld   de,(IBEG)
        xor  a
        ld   bc,(WINA)
        sbc  hl,de
        jr   nc,LET_PA       ;IF WINA>=IBEG
        ld   bc,(IBEG)       ;IF WINA< IBEG (window not fully open)
        dec  a

LET_PA: ld   (PINA),bc       ;PINA=adjusted LH window boundary
        ld   (WINDIFF),a
        ld   hl,(WINZ)
        and  a
        sbc  hl,bc
        inc  hl
        ld   b,h
        ld   c,l             ;BC=adjusted window width

        ;loop return point if window is fully open
        ;BC holds CPDR counter
ENLOOPZ:ld   hl,(CPOS)
        ld   de,(IEND)
        and  a
        sbc  hl,de
        jp   nc,EOINPUT       ;end of input

        ld   de,(CPOS)
        ld   (CNXT),de       ;lookahead pointer
        ld   de,0
        ld   (PBEG),de
        xor  a
        ld   (PLEN),a        ;PBEG,PLEN = matched start,length in window
        ld   hl,(WINZ)

        ld   a,(PWIDTHF)
        or   a
        jr   nz,FIND12
        inc  a               ;only done once
        ld   (PWIDTHF),a
        jr   LONGEST

FIND12:

        ;will search from the history match point, if found,
        ;otherwise from the RH window edge (WINZ), backwards
PA2PZ:  ld   de,(CPOS)
        ld   a,(de)          ;A =character at Coding Position
        cpdr                 ;search backwards for a match
        jp   z,MASTART
        ;no further matches in window; output the last longest match (if any)
        ;(PBEG) & (PLEN) will identify the last longest match found
LONGEST:;call OUTPUT

;******
        ;output subroutine
OUTPUT: ld   hl,(LPOS)
        ld   a,(PLEN)
        cp   1
        jr   z,MATCH1        ;IF PLEN=1
        jr   c,MATCH0        ;IF PLEN=0

;2+ matches: 2 bytes : llllllll oooooooo  : length+offset
MATCH2: ld   (hl),a          ;=(PLEN)
        inc  hl
        exx
        ld   hl,(CNXT)
        ld   bc,(PBEG)
        sbc  hl,bc
        ld   bc,(PLEN)       ;deduct (CNXT-CPOS);
        ld   b,0
        sbc  hl,bc           ;HL=lookback offset
        ld   a,l
        exx
        ld   (hl),a          ;=offset
        ld   c,2             ;C =byte copy counter
        dec  (ix+04)         ;=(PLEN)
        jr   OUTEND2

; 1 match:   1 byte  : mmmmmmmm           : mchar
MATCH1: ld   de,(PBEG)
        dec  (ix+04)         ;=(PLEN)
        jr   MATCHA

; 0 matches: 1 byte  : cccccccc           : cchar
MATCH0: ld   de,(CNXT)
MATCHA: ld   a,1             ;A =byte copy counter

OUTEND: ld   c,a
        ld   a,(de)
        ld   (hl),a
OUTEND2:ld   d,0
        ld   e,c
        ld   a,c

        exx
        ld   hl,(CODADD)     ;(CODADD)=coding flag byte address
        ld   c,(hl)
        sla  c
        dec  a
        add  a,c
        ld   (hl),a
        exx
        dec  (ix+30)         ;(CODSEQ)=coding flag byte sequence
        jr   nz,OUTADD

        inc  hl
        ld   (CODADD),hl     ;(CODADD)=coding flag byte address
        ld   (hl),d          ;D =0
        ld   (ix+30),8       ;(CODSEQ)=coding flag byte sequence
        inc  e

OUTADD: ld   hl,(OBYTES)
        add  hl,de           ;add DE bytes
        ld   (OBYTES),hl
        ld   hl,(LPOS)
        add  hl,de           ;add DE bytes
        ld   (LPOS),hl

        ld   de,(PLEN)       ;update pointers by PLEN+1
        ld   d,0
        inc  de
        ld   hl,(WINA)
        add  hl,de
        ld   (WINA),hl
        ld   hl,(PINA)
        add  hl,de
        ld   (PINA),hl
        ld   hl,(WINZ)
        add  hl,de
        ld   (WINZ),hl
        ld   hl,(CPOS)
        add  hl,de
        ld   (CPOS),hl
;******

        ld   a,(WINDIFF)
        or   a               ;window fully open?
        jp   nz,ENLOOP       ;no
        ld   bc,WWIDTH
        jp   ENLOOPZ

MASTART:push hl           ;1>;preserve match start registers
        exx               ;H'L' was saved prior to entering ENLOOP
        pop  hl           ;1<
        inc  hl
        ld   b,0             ;match length
        ld   de,(CPOS)
        ld   (PTMP),hl       ;match start

        ;determine the extent of the match
MALOOP: ld   a,(ILAST)
        cp   e
        jr   nz,MANEXT1
        ld   a,(ILAST+1)
        cp   d
        jr   z,NEXTSW        ;reached end of input

MANEXT1:ld   a,(CPOS)
        cp   l
        jr   nz,MANEXT2
        ld   a,(CPOS+1)
        cp   h
        jr   z,NEXTSW        ;reached Coding Position

MANEXT2:ld   a,(de)          ;compare lookahead character
        cp   (hl)            ;   with window character
        jr   nz,NEXTSW       ;no further match

        inc  b
        inc  hl              ;window pointer
        inc  de              ;lookahead pointer
        ld   a,b             ;A =new match length
        cp   (ix-01)         ;(MAXLEN)
        jr   nz,MALOOP

NEXTSW: ld   a,(PLEN)        ;A =longest match length so far
        cp   b
        ld   a,b             ;A =new match length
        jr   nc,NOTLONG

        ld   (CNXT),de
        ld   hl,(PTMP)
        ld   (PBEG),hl
        ld   (PLEN),a

NOTLONG:exx
        cp   (ix-01)         ;(MAXLEN)
        jp   nc,LONGEST
        ld   a,b
        or   c
        jp   nz,PA2PZ
        jp   LONGEST

EOINPUT:ld   hl,(CODADD)     ;(CODADD)=coding flag byte address
        ld   a,(hl)
        or   a
        jr   z,ENDRUN
EOINSLA:sla  a               ;left justify the final coding byte
        dec  (ix+30)         ;(CODSEQ)=coding flag byte sequence
        jr   nz,EOINSLA
        ld   (hl),a

        ;END OF THE ENCODING LOOP
ENDRUN: pop hl               ;restore H'L'
        exx

        call FPTOBC          ;BC=start of output
        ld   a,(ix+06)       ;(OBYTES)_lo
        ld   (bc),a          ;store the packed file size
        inc  bc              ; at the start of the file
        ld   a,(ix+07)       ;(OBYTES)_hi
        ld   (bc),a
        ld   hl,(STKEND)
        ld   de,5
        add  hl,de
        ld   (STKEND),hl     ;restore (STKEND)

        ret

BUFFER: defs MAXSIZE         ;workspace